/*
 * USB HID bootloader for PIC32 microcontroller.
 *
 * Based on Microchip sources.
 * Heavily rewritten by Serge Vakulenko, <serge@vak.ru>.
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the 'Company') for its PIC(R) Microcontroller is intended and
 * supplied to you, the Company's customer, for use solely and
 * exclusively on Microchip PIC Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN 'AS IS' CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 */
#include <string.h>
#include "usb_device.h"
#include "usb_function_hid.h"

#include <p32mx/p32mx_core.h>
#include <p32mx/p32mx_wdt.h>
#include <p32mx/p32mx_nvm.h>
#include <p32mx/p32mx_usbotg.h>
#include <p32mx/p32mx_port.h>
#include <p32mx/p32mx_dmac.h>
#include <p32mx/p32mx_sys.h>
#include <p32mx/bootloader.h>

/*
 * Commands of bootloader protocol.
 */
#define	QUERY_DEVICE	    0x02    // Command that the host uses to learn
                                    // about the device (what regions can be
                                    // programmed, and what type of memory
                                    // is the region)
#define	UNLOCK_CONFIG	    0x03    // Note, this command is used for both
                                    // locking and unlocking the config bits
#define ERASE_DEVICE	    0x04    // Host sends this command to start
                                    // an erase operation.  Firmware controls
                                    // which pages should be erased.
#define PROGRAM_DEVICE	    0x05    // If host is going to send a full
                                    // REQUEST_SIZE to be programmed, it
                                    // uses this command.
#define	PROGRAM_COMPLETE    0x06    // If host send less than a REQUEST_SIZE
                                    // to be programmed, or if it wished
                                    // to program whatever was left in
                                    // the buffer, it uses this command.
#define GET_DATA	    0x07    // The host sends this command in order
                                    // to read out memory from the device.
                                    // Used during verify (and read/export
                                    // hex operations)
#define	RESET_DEVICE	    0x08    // Resets the microcontroller, so it can
                                    // update the config bits (if they were
                                    // programmed, and so as to leave
                                    // the bootloader (and potentially go back
                                    // into the main application)

/*
 * HID packet structure.
 */
#define	PACKET_SIZE	    64      // HID packet size
#define REQUEST_SIZE        56      // Number of bytes of a standard request

typedef union __attribute__ ((packed))
{
    unsigned char Contents [PACKET_SIZE];

    struct __attribute__ ((packed)) {
        unsigned char Command;
        unsigned Address;
        unsigned char Size;
        unsigned char PadBytes [PACKET_SIZE - 6 - REQUEST_SIZE];
        unsigned int Data [REQUEST_SIZE / sizeof(unsigned)];
    };

    struct __attribute__ ((packed)) {
        unsigned char Command;
        unsigned char PacketDataFieldSize;
        unsigned char DeviceFamily;
        unsigned char Type1;
        unsigned long Address1;
        unsigned long Length1;
        unsigned char Type2;
        unsigned long Address2;
        unsigned long Length2;
        unsigned char Type3;
        unsigned long Address3;
        unsigned long Length3;
        unsigned char Type4;		// End of sections list indicator goes here, fill with 0xFF.
        unsigned char ExtraPadBytes[33];
    };
} packet_t;

static packet_t send;           // 64-byte send buffer (EP1 IN to the PC)
static packet_t receive_buffer; // 64-byte receive buffer (EP1 OUT from the PC)
static packet_t receive;        // copy of receive buffer, for processing

static USB_HANDLE request_handle;

/*
 * Placed at fixed address 0x80000000 by linker.
 * To enter boot mode from user program, set this value
 * to 0x12345678 and perform a software reset.
 */
unsigned int reset_key
     __attribute__((section(".sdata")));

static unsigned int buf [32];
static unsigned int buf_index;
static unsigned int base_address;

/*
 * Chip configuration.
 */
#define _BUILD_CONFIG
#include <p32mx/p32mx_devcfg.h>
#ifdef CRYSTAL_12MHZ
    #define CONFIG_FPLLIDIV             IDIV_3
    #define CONFIG_UPLLIDIV             IDIV_3
#elif defined(CRYSTAL_16MHZ)
    #define CONFIG_FPLLIDIV             IDIV_4
    #define CONFIG_UPLLIDIV             IDIV_4
#endif
   #define CONFIG_BWP      0  /* Protected */
   //#define CONFIG_CP       0  /* Protected */
#if defined(__32MX2XX__)
   #define CONFIG_PWP      PWP_SIZE(8) /* lower 8KB is Protected */
#else
   #undef  BOOT_SECTION /* all text move to boot flash */
   #define BOOT_SECTION 
#endif

   #define CONFIG_JTAGEN   1 /* Enabled : disabled in main */
#if defined(PINGUINO_MX220)
   #define CONFIG_ICSEL    ICS_PGx3  /* PGEx3 */
#elif defined(DIP_MX220)
   #define CONFIG_ICSEL    ICS_PGx3  /* PGEx3 */
#endif

#include <p32mx/p32mx_devcfg_build.h>
/*
 * Boot Loader configuration.
 */
#include <p32mx/bootloader.h>

#ifndef APP_FLASH_START
#define APP_FLASH_START	BLCFG0
#endif
#ifndef APP_FLASH_END
#define APP_FLASH_END	(BLCFG1 -1)
#endif
#ifndef APP_START_ADDR
#define APP_START_ADDR	BLCFG2
#endif

  #ifndef FLASH_START_DEFAULT
     #if defined(__32MX2XX__)
        #define FLASH_START_DEFAULT  0xbd002000
     #else
        #define FLASH_START_DEFAULT  0xbd000000
     #endif
  #endif

  BLCFG0_INIT(FLASH_START_DEFAULT);

  #ifndef FLASH_END_DEFAULT
     #if defined(__32MX2XX__)
        #define FLASH_END_DEFAULT  0xbd020000 // 128K
     #else
        #define FLASH_END_DEFAULT  0xbd080000 // 512K
     #endif
  #endif
  BLCFG1_INIT(FLASH_END_DEFAULT);

  #ifndef START_ADDR_DEFAULT  /* search from APP_FLASH_START */
  #define START_ADDR_DEFAULT  0xffffffff 
  #endif
  BLCFG2_INIT(START_ADDR_DEFAULT);

/*
 * Flash memory.
 */
#define FLASH_PAGESZ        4096        /* bytes per page, for erasing */


/* LED_TYPE: 0: LAT_SET  (to light on / LAT_CLR to light off)
             1: LAT_CLR  (reversed logic of 0) 
             2: TRIS_SET (reversed logic of 3) 
             3: TRIS_CLR (to light on / TRIS_SET to light off)
   BUT_TYPE  0: PULLUP (2xx only)
             1: NO PULLUP
 */
#if defined(PINGUINO_MX220)
  BLCFG3_INIT(  .REDLED_PORT = PORTA, .REDLED_BIT = 10,
		.GRNLED_PORT = PORTB, .GRNLED_BIT = 15,
		.BUT_PORT    = PORTB,    .BUT_BIT = 7,
		.REDLED_TYPE = 2, .GRNLED_TYPE = 3, .BUT_TYPE = 0 );
#elif defined(DIP_MX220)
  BLCFG3_INIT(  .REDLED_PORT = PORTB, .REDLED_BIT = 14,
		.GRNLED_PORT = PORTB, .GRNLED_BIT = 15,
		.BUT_PORT    = PORTB,    .BUT_BIT = 7,
		.REDLED_TYPE = 2, .GRNLED_TYPE = 3, .BUT_TYPE = 0 );
#elif defined(PINGUINO_MICRO)
  BLCFG3_INIT(  .REDLED_PORT = PORTD, .REDLED_BIT = 1,
		.GRNLED_PORT = PORTG, .GRNLED_BIT = 6,
		.BUT_PORT    = PORTD,    .BUT_BIT = 0,
		.REDLED_TYPE = 2, .GRNLED_TYPE = 3, .BUT_TYPE = 0 );
#else
  BLCFG3_INIT(  .REDLED_PORT = PORTB, .REDLED_BIT = 14,
		.GRNLED_PORT = PORTB, .GRNLED_BIT = 15,
		.BUT_PORT    = PORTB,    .BUT_BIT = 7,
		.REDLED_TYPE = 2, .GRNLED_TYPE = 3, .BUT_TYPE = 0 );
#endif

#ifndef CFG_REDLED_TYPE
#define CFG_REDLED_TYPE (((struct blcfg3 *)_BLCFG3)-> REDLED_TYPE)
#endif
#ifndef CFG_REDLED_PORT
#define CFG_REDLED_PORT (((struct blcfg3 *)_BLCFG3)-> REDLED_PORT)
#endif
#ifndef CFG_REDLED_BIT
#define CFG_REDLED_BIT  (((struct blcfg3 *)_BLCFG3)-> REDLED_BIT)
#endif

#ifndef CFG_GRNLED_TYPE
#define CFG_GRNLED_TYPE (((struct blcfg3 *)_BLCFG3)-> GRNLED_TYPE)
#endif
#ifndef CFG_GRNLED_PORT
#define CFG_GRNLED_PORT (((struct blcfg3 *)_BLCFG3)-> GRNLED_PORT)
#endif
#ifndef CFG_GRNLED_BIT
#define CFG_GRNLED_BIT  (((struct blcfg3 *)_BLCFG3)-> GRNLED_BIT)
#endif

#ifndef CFG_BUT_TYPE
#define CFG_BUT_TYPE (((struct blcfg3 *)_BLCFG3)-> BUT_TYPE)
#endif
#ifndef CFG_BUT_PORT
#define CFG_BUT_PORT (((struct blcfg3 *)_BLCFG3)-> BUT_PORT)
#endif
#ifndef CFG_BUT_BIT
#define CFG_BUT_BIT  (((struct blcfg3 *)_BLCFG3)-> BUT_BIT)
#endif

#if 1
struct led_ctrl {
	volatile uint16_t *tris_reg;
	volatile uint16_t *lat_reg;
	volatile uint16_t *on_reg;
	volatile uint16_t *off_reg;
	uint16_t mask;
};
static struct led_ctrl grn, red;

static inline void connect_led_on() {
	*grn.on_reg = grn.mask;
}
static inline void connect_led_off() {
	*grn.off_reg = grn.mask;
}
static inline void busy_led_on() {
	*red.on_reg = red.mask;
}
static inline void busy_led_off() {
	*red.off_reg = red.mask;
}
static inline void port_term() {
	*(red.tris_reg +2) = red.mask; // TRIS_SET
	*red.lat_reg = red.mask;
	*(grn.tris_reg +2) = grn.mask; // TRIS_SET
	*grn.lat_reg = grn.mask;
#if defined(__32MX1XX__) || defined(__32MX2XX__)
	regx_clr(CNPU, CFG_BUT_PORT, CFG_BUT_BIT);
#endif
}

static inline void port_init() {
    int i, flag;
    struct led_ctrl *led;
    regx_set(TRIS, CFG_BUT_PORT, CFG_BUT_BIT);
#if defined(__32MX1XX__) || defined(__32MX2XX__)
     if (! CFG_BUT_TYPE ) regx_set(CNPU, CFG_BUT_PORT, CFG_BUT_BIT);
#endif
    grn.mask = _BV(CFG_GRNLED_BIT);
    grn.tris_reg = & TRIS_CLR(CFG_GRNLED_PORT);
  
    red.mask = _BV(CFG_REDLED_BIT);
    red.tris_reg = & TRIS_CLR(CFG_REDLED_PORT);

    flag = CFG_GRNLED_TYPE;
    led = &grn;
    for (i=0; i<2; i++) {
        led->lat_reg =  led->tris_reg + 16;
	if (flag & 1) {
		led->on_reg = led->tris_reg;
	} else {
		led->on_reg = led->lat_reg;
	}
	if (flag & 2) {
		led->off_reg = led->on_reg + 2;
	} else {
		led->off_reg = led->on_reg;
		led->on_reg += 2;
	}
	if (flag & 3) {
		*(led->tris_reg +2) = led->mask; // TRIS_SET
	} else {
		*led->tris_reg  = led->mask; // TRIS_CLR
	}
    	flag = CFG_REDLED_TYPE;
    	led = &red;
    }
}
#else

static inline void port_init() {
    regx_set(TRIS, CFG_BUT_PORT, CFG_BUT_BIT);
#if defined(__32MX1XX__) || defined(__32MX2XX__)
     if (! CFG_BUT_TYPE ) regx_set(CNPU, CFG_BUT_PORT, CFG_BUT_BIT);
#endif
    switch (CFG_REDLED_TYPE) {
    case 0:
    	regx_set(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_clr(LAT,  CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 1:
    	regx_clr(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_set(LAT,  CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 2:
    	regx_clr(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_clr(LAT,  CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 3:
    	regx_clr(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_clr(LAT,  CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    }
    switch (CFG_GRNLED_TYPE) {
    case 0:
    	regx_clr(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 1:
    	regx_clr(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 2:
    	regx_clr(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
    	regx_set(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 3:
    	regx_set(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    }
}

static inline void port_term() {
    	regx_set(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_clr(LAT,  CFG_REDLED_PORT, CFG_REDLED_BIT);
    	regx_set(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
}

static inline void connect_led_on() {
    switch (CFG_GRNLED_TYPE) {
    case 0:
    	regx_set(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 1:
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 2:
    	regx_set(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 3:
    	regx_clr(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    }
}

static inline void connect_led_off() {
    switch (CFG_GRNLED_TYPE) {
    case 0:
    	regx_clr(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 1:
    	regx_set(LAT,  CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 2:
    	regx_clr(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    case 3:
    	regx_set(TRIS, CFG_GRNLED_PORT, CFG_GRNLED_BIT);
	break;
    }
}

static inline void busy_led_on() {
    switch (CFG_REDLED_TYPE) {
    case 0:
    	regx_set(LAT, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 1:
    	regx_clr(LAT, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 2:
    	regx_set(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 3:
    	regx_clr(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    }
}

static inline void busy_led_off() {
    switch (CFG_REDLED_TYPE) {
    case 0:
    	regx_clr(LAT, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 1:
    	regx_set(LAT, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 2:
    	regx_clr(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    case 3:
    	regx_set(TRIS, CFG_REDLED_PORT, CFG_REDLED_BIT);
	break;
    }
}
#endif

static inline int button_pressed()
{
    return regx_is_clr(PORT, CFG_BUT_PORT, CFG_BUT_BIT);
}

/*
 * Reset the microrontroller.
 */
static void BOOT_SECTION soft_reset()
{
    cli();

    if (! (DMACON & 0x1000)) {
        DMACON_SET = 0x1000;
        while (DMACON & 0x800)
            continue;
    }
    SYSKEY = 0;
    SYSKEY = 0xaa996655;
    SYSKEY = 0x556699aa;

    RSWRST_SET = 1;
    (void) RSWRST;

    for (;;)
        continue;
}

/*
 * Microsecond delay routine for MIPS processor.
 */
/*
 * Clear an array.
 */
static inline void memzero (void *data, unsigned nbytes)
{
   memset(data, 0, nbytes);
}

static inline void memcopy (void *from, void *to, unsigned nbytes)
{
   memcpy(to, from, nbytes);
}

/*
 * Perform non-volatile memory operation.
 */

/*
 * Use word writes to write code chunks less than a full 64 byte block size.
 */
static void BOOT_SECTION write_flash_block()
{
    unsigned int i;
    unsigned int addr;

    for (i = 0; buf_index > 0; buf_index--) {
        // Write word to flash memory.
        addr = base_address - buf_index * sizeof(unsigned);
        addr |= 0xc0000000; // to kseg1
	if ((APP_FLASH_START <= addr) && (addr < APP_FLASH_END +1))
            __nvm_program_word (addr, buf [i++]);
    }
}

/*
 * A command packet was received from PC.
 * Process it and return 1 when a reply is ready.
 */
static int BOOT_SECTION handle_packet()
{
    int i;
    unsigned addr;

    switch (receive.Command) {
    case QUERY_DEVICE:
        // Prepare a response packet, which lets the PC software know
        // about the memory ranges of this device.
        memzero (&send, PACKET_SIZE);
        send.Command = QUERY_DEVICE;
        send.PacketDataFieldSize = REQUEST_SIZE;
        send.DeviceFamily = 3;            /* PIC32 */
        send.Type1 = 1;                   /* 'program' memory type */
        send.Address1 = APP_FLASH_START & ~ 0xc0000000;
        send.Length1 = APP_FLASH_END + 1 - APP_FLASH_START;
        send.Type2 = 0xFF;                /* end of list */
        return 1;

    case UNLOCK_CONFIG:
        return 0;

    case ERASE_DEVICE:
	addr = APP_FLASH_START;
	while (addr < APP_FLASH_END + 1) {
            /* Erase flash page */
            __nvm_page_erase (addr);
            // Call this function periodically to prevent falling
            // off the bus if any SETUP packets should happen to arrive.
            usb_device_tasks();
	    addr += FLASH_PAGESZ;
        }
        // Good practice to clear WREN bit anytime we are
        // not expecting to do erase/write operations,
        // further reducing probability of accidental activation.
        NVMCON_CLR = _BV(NVMCON_WREN);
        return 0;

    case PROGRAM_DEVICE:
        if (base_address == 0)
            base_address = receive.Address;

        if (base_address == receive.Address) {
            for (i = 0; i < receive.Size/sizeof(unsigned); i++) {
                unsigned int index;

                index = (REQUEST_SIZE - receive.Size) / sizeof(unsigned) + i;

                // Data field is right justified.
                // Need to put it in the buffer left justified.
                buf [buf_index++] =
                    receive.Data [(REQUEST_SIZE - receive.Size) / sizeof(unsigned) + i];
                base_address += sizeof(unsigned);
                if (buf_index == REQUEST_SIZE / sizeof(unsigned)) {
                    write_flash_block();
                }
            }
        }
        // else host sent us a non-contiguous packet address...
        // to make this firmware simpler, host should not do this without
        // sending a PROGRAM_COMPLETE command in between program sections.
        return 0;

    case PROGRAM_COMPLETE:
        write_flash_block();

        // Reinitialize pointer to an invalid range, so we know the next
        // PROGRAM_DEVICE will be the start address of a contiguous section.
        base_address = 0;
        return 0;

    case GET_DATA:
        memzero (&send, PACKET_SIZE);
        send.Command = GET_DATA;
        send.Address = receive.Address;
        send.Size = receive.Size;
        memcopy ((void*) (receive.Address | 0x80000000),
            send.Data + (REQUEST_SIZE - receive.Size)/sizeof(unsigned),
            receive.Size);
        return 1;

    case RESET_DEVICE:
        // Disable USB module
        U_CON = 0x0000;

        // And wait awhile for the USB cable capacitance to discharge
        // down to disconnected (SE0) state.   Otherwise host might not
        // realize we disconnected/reconnected when we do the reset.
        delay_us (1000);

        soft_reset();
        return 0;
    }
    return 0;
}

/*
 * Main program entry point.
 */
int BOOT_SECTION main()
{
#if 0 // intialized in crt0
    /* Initialize STATUS register: master interrupt disable. */
    write_c0_status (ST_CU0 | ST_BEV);

    /* Clear CAUSE register: use special interrupt vector 0x200. */
    write_c0_cause (CA_IV);
#endif

#if 0
    /*
     * Setup wait states.
     */
    CHECON = 2;
    BMXCON_CLR = 0x40;
    CHECON_SET = 0x30;
#endif

#ifndef DDPCON
  #define DDPCON	CFGCON
#endif
    /* Disable JTAG port, to use it for i/o. */
    DDPCON = 0;

#if 0 // intialized in crt0
    /* Config register: enable kseg0 caching. */
    write_c0_status (read_c0_status() | 3);
#endif

    //cinit();
    //cputs("=0=\n");

    // Initialize all of the push buttons
    // Initialize all of the LED pins
    port_init();

    /*
     * To call the normal application (i.e. don't go into the bootloader)
     * we need to make sure that the PRG button is not pressed, and that we
     * don't have a software reset.
     */
    if (! button_pressed() &&
        (! (RCON & _BV(RCON_SWR)) || reset_key != 0x12345678))
    {
        void (*fptr)(void);
        unsigned entry;
        // Must clear out software key
	reset_key = 0;

	entry = APP_START_ADDR;
	if (!entry || (entry == 0xffffffff)) {
		entry = APP_FLASH_START;
		while (*(uint32_t *)entry != 0x3c1a9d00){ /*   lui  k0,0x9d00 */
			entry += 0x1000;
			if (entry > APP_FLASH_END + 1) break;
		}
	}
        // If there is NO real program to execute, then fall through to the bootloader
	if ((entry < APP_FLASH_END + 1) && (*(uint32_t *)entry != 0xffffffff)) {
            fptr = (void (*)(void)) entry;
	    port_term();
            fptr();
            for (;;)
                continue;
	}
    }
    // Must clear out software key.
    reset_key = 0;
    RCON_CLR = _BV(RCON_SWR);

    // Initialize USB module SFRs and firmware variables to known states.
    usb_device_init();

    USB_HANDLE reply_handle = 0;
    int packet_received = 0;
    for (;;) {
	// Check bus status and service USB interrupts.
        usb_device_tasks();

        // User Application USB tasks
        if (usb_device_state < CONFIGURED_STATE || usb_is_device_suspended()) {
		connect_led_off();
		continue;
	} else {
		connect_led_on();
	}

	busy_led_on();
        // Are we done sending the last response.  We need to be before we
        // receive the next command because we clear the send buffer
        // once we receive a command
        if (reply_handle != 0) {
            if (usb_handle_busy (reply_handle))
                continue;
            reply_handle = 0;
        }

        if (packet_received) {
            if (handle_packet())
                reply_handle = usb_transfer_one_packet (HID_EP,
                    IN_TO_HOST, &send.Contents[0], PACKET_SIZE);
            packet_received = 0;
	    busy_led_off();
            continue;
        }

        // Did we receive a command?
        if (usb_handle_busy (request_handle))
            continue;

        // Make a copy of received data.
        memcopy (&receive_buffer, &receive, PACKET_SIZE);
        packet_received = 1;

        // Restart receiver, to be ready for a next packet.
        request_handle = usb_transfer_one_packet (HID_EP, OUT_FROM_HOST,
            (unsigned char*) &receive_buffer, PACKET_SIZE);
	busy_led_off();
    }
}

/*
 * USB Callback Functions
 */
/*
 * Process device-specific SETUP requests.
 */
void BOOT_SECTION usbcb_check_other_req()
{
   hid_check_request();
   __nop();
   return;
}

/*
 * This function is called when the device becomes initialized.
 * It should initialize the endpoints for the device's usage
 * according to the current configuration.
 */
void BOOT_SECTION usbcb_init_ep()
{
    // Enable the HID endpoint
    usb_enable_endpoint (HID_EP, USB_IN_ENABLED | USB_OUT_ENABLED |
        USB_HANDSHAKE_ENABLED | USB_DISALLOW_SETUP);

    // Arm the OUT endpoint for the first packet
    request_handle = usb_rx_one_packet (HID_EP,
        &receive_buffer.Contents[0], PACKET_SIZE);
}

/*
 * The device descriptor.
 */
const USB_DEVICE_DESCRIPTOR usb_device = {
    sizeof (usb_device),    // Size of this descriptor in bytes
    USB_DESCRIPTOR_DEVICE,  // DEVICE descriptor type
    0x0200,                 // USB Spec Release Number in BCD format
    0x00,                   // Class Code
    0x00,                   // Subclass code
    0x00,                   // Protocol code
    USB_EP0_BUFF_SIZE,      // Max packet size for EP0
    0x04D8,                 // Vendor ID
    0x003C,                 // Product ID: USB HID Bootloader
    0x0002,                 // Device release number in BCD format
    1,                      // Manufacturer string index
    2,                      // Product string index
    0,                      // Device serial number string index
    1,                      // Number of possible configurations
};

/*
 * Configuration 1 descriptor
 */
const unsigned char usb_config1_descriptor[] = {
    /*
     * Configuration descriptor
     */
    sizeof (USB_CONFIGURATION_DESCRIPTOR),
    USB_DESCRIPTOR_CONFIGURATION,
    0x29, 0x00,             // Total length of data for this cfg
    1,                      // Number of interfaces in this cfg
    1,                      // Index value of this configuration
    0,                      // Configuration string index
    _DEFAULT | _SELF,       // Attributes
    50,                     // Max power consumption (2X mA)

    /*
     * Interface descriptor
     */
    sizeof (USB_INTERFACE_DESCRIPTOR),
    USB_DESCRIPTOR_INTERFACE,
    0,                      // Interface Number
    0,                      // Alternate Setting Number
    2,                      // Number of endpoints in this intf
    HID_INTF,               // Class code
    0,                      // Subclass code
    0,                      // Protocol code
    0,                      // Interface string index

    /*
     * HID Class-Specific descriptor
     */
    sizeof (USB_HID_DSC) + 3,
    DSC_HID,
    0x11, 0x01,             // HID Spec Release Number in BCD format (1.11)
    0x00,                   // Country Code (0x00 for Not supported)
    HID_NUM_OF_DSC,         // Number of class descriptors
    DSC_RPT,                // Report descriptor type
    HID_RPT01_SIZE, 0x00,   // Size of the report descriptor

    /*
     * Endpoint descriptor
     */
    sizeof (USB_ENDPOINT_DESCRIPTOR),
    USB_DESCRIPTOR_ENDPOINT,
    HID_EP | _EP_IN,        // EndpointAddress
    _INTERRUPT,             // Attributes
    PACKET_SIZE, 0,         // Size
    1,                      // Interval

    /*
     * Endpoint descriptor
     */
    sizeof (USB_ENDPOINT_DESCRIPTOR),
    USB_DESCRIPTOR_ENDPOINT,
    HID_EP | _EP_OUT,       // EndpointAddress
    _INTERRUPT,             // Attributes
    PACKET_SIZE, 0,         // Size
    1,                      // Interval
};

/*
 * Class specific descriptor - HID
 */
const unsigned char hid_rpt01 [HID_RPT01_SIZE] = {
    0x06, 0x00, 0xFF,       // Usage Page = 0xFF00 (Vendor Defined Page 1)
    0x09, 0x01,             // Usage (Vendor Usage 1)
    0xA1, 0x01,             // Collection (Application)

    0x19, 0x01,             // Usage Minimum
    0x29, 0x40,             // Usage Maximum 64 input usages total (0x01 to 0x40)
    0x15, 0x00,             // Logical Minimum (data bytes in the report may have minimum value = 0x00)
    0x26, 0xFF, 0x00,       // Logical Maximum (data bytes in the report may have maximum value = 0x00FF = unsigned 255)
    0x75, 0x08,             // Report Size: 8-bit field size
    0x95, 0x40,             // Report Count: Make sixty-four 8-bit fields (the next time the parser hits an "Input", "Output", or "Feature" item)
    0x81, 0x00,             // Input (Data, Array, Abs): Instantiates input packet fields based on the above report size, count, logical min/max, and usage.

    0x19, 0x01,             // Usage Minimum
    0x29, 0x40,             // Usage Maximum 64 output usages total (0x01 to 0x40)
    0x91, 0x00,             // Output (Data, Array, Abs): Instantiates output packet fields.  Uses same report size and count as "Input" fields, since nothing new/different was specified to the parser since the "Input" item.

    0xC0,                   // End Collection
};

/*
 * USB Strings
 */
static const USB_STRING_INIT(1) string0_descriptor = {
    sizeof(string0_descriptor),
    USB_DESCRIPTOR_STRING,              /* Language code */
    { 0x0409 },
};

static const USB_STRING_INIT(25) string1_descriptor = {
    sizeof(string1_descriptor),
    USB_DESCRIPTOR_STRING,              /* Manufacturer */
    { 'M','i','c','r','o','c','h','i','p',' ',
      'T','e','c','h','n','o','l','o','g','y',
      ' ','I','n','c','.' },
};

static const USB_STRING_INIT(18) string2_descriptor = {
    sizeof(string2_descriptor),
    USB_DESCRIPTOR_STRING,              /* Product */
    { 'U','S','B',' ','H','I','D',' ','B','o',
      'o','t','l','o','a','d','e','r' },
};

/*
 * Array of configuration descriptors
 */
const unsigned char *const usb_config[] = {
    (const unsigned char *const) &usb_config1_descriptor,
};

/*
 * Array of string descriptors
 */
const unsigned char *const usb_string[USB_NUM_STRING_DESCRIPTORS] = {
    (const unsigned char *const) &string0_descriptor,
    (const unsigned char *const) &string1_descriptor,
    (const unsigned char *const) &string2_descriptor,
};
